#include <aalib.h>
#include <stdlib.h>
#include <stdio.h>

aa_context *context = NULL;
unsigned char *buffer;
aa_palette palette;
aa_renderparams *params;
void (*kbd_handler) (int scancode, int press) = NULL;
static int scantokey[128];
unsigned char scanpressed[128]={0,};
int version=0;

struct vga_modeinfo {
	int width;
	int height;
	int bytesperpixel;
	int colors;
	int linewidth;		/* scanline width in bytes */
	int maxlogicalwidth;	/* maximum logical scanline width */
	int startaddressrange;	/* changeable bits set */
	int maxpixels;		/* video memory / bytesperpixel */
	int haveblit;		/* mask of blit functions available */
	int flags;		/* other flags */

	/* Extended fields: */

	int chiptype;		/* Chiptype detected */
	int memory;		/* videomemory in KB */
	int linewidth_unit;	/* Use only a multiple of this as parameter for set_logicalwidth and
				 * set_displaystart */
	char *linear_aperture;	/* points to mmap secondary mem aperture of card (NULL if
				 * unavailable) */
	int aperture_size;	/* size of aperture in KB if size>=videomemory. 0 if unavail */
	void (*set_aperture_page) (int page);
	/* if aperture_size<videomemory select a memory page */
	void *extensions;	/* points to copy of eeprom for mach32 */
	/* depends from actual driver/chiptype.. etc. */
} mode = {

	320, 200, 1, 256, 320, 320, 0, 64000, 0, 0, 0, 64000, 0, NULL, 0, NULL, NULL
};

int vga_init()
{
	return 1;
}

int vga_setmode(int x)
{
	if (x == 0 && context)
		return aa_close(context), free(buffer), 0;
	if (x == 5 && !context)
		if (context = aa_autoinit(&aa_defparams)) {
			if (buffer = malloc(64000)) {
				params = aa_getrenderparams();
				params->bright = 100;
				params->contrast = 100;
				memset(aa_image(context), 0, aa_imgwidth(context) * aa_imgheight(context));
				fprintf(stderr,"%d,%d\n",aa_imgwidth(context),aa_imgheight(context));
				if (aa_imgwidth(context)>=320)
					version=10;
				else
					version=20;
				if (aa_imgheight(context)>=200)
					version+=1;
				else if (aa_imgheight(context)>=100)
					version+=2;
				else if (aa_imgheight(context)>=66)
					version+=3;
				else
					version+=4;
				return 0;
			} else
				exit((perror("malloc"), -1));
		}
	return fprintf(stderr, "aavga: vga_setmode(%d) failed!\n", x), 1;
}

void *vga_getgraphmem()
{
	return buffer;
}

void vga_setpage(int i)
{
	int y, x;
	unsigned char *img = aa_image(context);

	if (i) 
		fprintf(stderr, "vga_setpage(%d) called...%d!=0!!\n", x, x);

	switch(version) {
		case 11:
			for (y = 0; y < 200; y++)
				memcpy(img+y*aa_imgwidth(context),buffer+y*320,320);
			break;
		case 12:
			for (y=0;y<100;y++)
				memcpy(img+y*aa_imgwidth(context),buffer+y*320*2,320);
			break;
		case 13:
			for (y=0;y<66;y++)
				memcpy(img+y*aa_imgwidth(context),buffer+y*320*3,320);
			break;
		case 14:
			for (y=0;y<50;y++)
				memcpy(img+y*aa_imgwidth(context),buffer+y*320*4,320);
			break;
		case 21:
			for (y = 0; y < 200; y++)
				for (x=0;x<160;x++)
					img[y*aa_imgwidth(context)+x]=buffer[y*320+x*2];
			break;
		case 22:
			for (y=0;y<100;y++)
				for (x=0;x<160;x++)
					img[y*aa_imgwidth(context)+x]=buffer[y*2*320+x*2];
			break;
		case 23:
			for (y=0;y<66;y++)
				for (x=0;x<160;x++)
					img[y*aa_imgwidth(context)+x]=buffer[y*3*320+x*2];
			break;
		case 24:
			for (y=0;y<50;y++)
				for (x=0;x<160;x++)
					img[y*aa_imgwidth(context)+x]=buffer[y*4*320+x*2];
			break;
		default:
			fprintf(stderr,"uh, version = %d!\n",version);
	}		
	aa_renderpalette(context, palette, params, 0, 0, aa_scrwidth(context), aa_scrheight(context));
	aa_flush(context);
}

int vga_lastmodenumber()
{
	return 5;
}

int vga_hasmode(int x)
{
	return (x == 5);
}

struct vga_modeinfo *vga_getmodeinfo(int x)
{
	return &mode;
}

int vga_getmodenumber(char *s)
{
	return 5;
}

int vga_getcolors()
{
	return 256;
}

int vga_setpalvec(int a, int b, int *n)
{
	int i;

	for (i = a; i < b; i++)
		aa_setpalette(palette, i, n[i * 3], n[i * 3 + 1], n[i * 3 + 2]);
}

void vga_waitretrace()
{
}

int vga_oktowrite()
{
	return 1;
}

int vga_getmousetype()
{
	return 0;
}

void vga_runinbackground(int x)
{
}

int mouse_init()
{
	return 1;
}

void mouse_seteventhandler(void *h)
{
}

void mouse_close()
{
}

int mouse_update()
{
	return 0;
}

int keyboard_init()
{
	int i;

	if (!context)
		return fprintf(stderr, "Hrm, called before vga_setmode()?\n"), 1;
		
	for (i = 0; i < 128; i++)
		scantokey[i] = ' ';

	scantokey[59] = AA_UNKNOWN;		/* F1 */ // row 0

	scantokey[60] = AA_UNKNOWN;
	scantokey[61] = AA_UNKNOWN;
	scantokey[62] = AA_UNKNOWN;
	scantokey[63] = AA_UNKNOWN;
	scantokey[64] = AA_UNKNOWN;
	scantokey[65] = AA_UNKNOWN;
	scantokey[66] = AA_UNKNOWN;
	scantokey[67] = AA_UNKNOWN;
	scantokey[68] = AA_UNKNOWN;
	scantokey[87] = AA_UNKNOWN;
	scantokey[88] = AA_UNKNOWN;		/* F12 */

	scantokey[1] = AA_UNKNOWN; /* escape */ 	// row 1

	scantokey[2] = '1';
	scantokey[3] = '2';
	scantokey[4] = '3';
	scantokey[5] = '4';
	scantokey[6] = '5';
	scantokey[7] = '6';
	scantokey[8] = '7';
	scantokey[9] = '8';
	scantokey[10] = '9';
	scantokey[11] = '0';
	scantokey[12] = '-';
	scantokey[13] = '=';
	scantokey[14] = AA_BACKSPACE;

	scantokey[15] = '\t';	// row 2

	scantokey[16] = 'q';
	scantokey[17] = 'w';
	scantokey[18] = 'e';
	scantokey[19] = 'r';
	scantokey[20] = 't';
	scantokey[21] = 'y';
	scantokey[22] = 'u';
	scantokey[23] = 'i';
	scantokey[24] = 'o';
	scantokey[25] = 'p';
	scantokey[26] = '[';
	scantokey[27] = ']';
	scantokey[28] = '\r';		/* ENTER */

	scantokey[30] = 'a';	// row 3

	scantokey[31] = 's';
	scantokey[32] = 'd';
	scantokey[33] = 'f';
	scantokey[34] = 'g';
	scantokey[35] = 'h';
	scantokey[36] = 'j';
	scantokey[37] = 'k';
	scantokey[38] = 'l';
	scantokey[39] = ';';
	scantokey[40] = '\'';
	scantokey[41] = '`';

	scantokey[42] = AA_UNKNOWN;	/* shift */ // row 4

	scantokey[43] = '\\';
	scantokey[44] = 'z';
	scantokey[45] = 'x';
	scantokey[46] = 'c';
	scantokey[47] = 'v';
	scantokey[48] = 'b';
	scantokey[49] = 'n';
	scantokey[50] = 'm';
	scantokey[51] = ',';
	scantokey[52] = '.';
	scantokey[53] = '/';
	scantokey[54] = AA_UNKNOWN;	/* shift */

	scantokey[29] = AA_UNKNOWN;	// row 5 /* ctrl */

	scantokey[56] = AA_UNKNOWN;	/* alt */
	scantokey[57] = ' ';
	scantokey[100] = AA_UNKNOWN;	/* alt */
	scantokey[97] = AA_UNKNOWN;	/* ctrl */


	scantokey[98] = '/';
	scantokey[55] = '8';
	scantokey[74] = '-';
	scantokey[71] = AA_UNKNOWN;	/* home */
	scantokey[72] = AA_UP;
	scantokey[73] = AA_UNKNOWN;	/* pgup */
	scantokey[75] = AA_LEFT;
	scantokey[76] = '5';
	scantokey[77] = AA_RIGHT;
	scantokey[78] = '+';
	scantokey[79] = AA_UNKNOWN;	/* end */
	scantokey[80] = AA_DOWN;
	scantokey[81] = AA_UNKNOWN;	/* pgdn */
	scantokey[82] = AA_UNKNOWN;	/* ins */
	scantokey[83] = AA_UNKNOWN;	/* del */
	scantokey[96] = '\n';

	scantokey[103] = AA_UP;		// arrow pad

	scantokey[108] = AA_DOWN;
	scantokey[105] = AA_LEFT;
	scantokey[106] = AA_RIGHT;

	scantokey[119] = AA_UNKNOWN; /* pause */	// misc

	if (!aa_autoinitkbd(context, AA_SENDRELEASE))
		return fprintf(stderr, "Error in aa_autoinitkbd!\n"), 1;

	return 0;
}

int keyboard_update()
{
	int i,stat=1,scan;
	if ((i=aa_getevent(context,0))==AA_NONE)
		return 0;
	if (i>=AA_UNKNOWN && i<AA_RELEASE) {
		fprintf(stderr,"key: %d pressed\n");
		return 1;
	}
	if (i>=AA_RELEASE) {
		stat=0;
		i&=~AA_RELEASE;
	}
	for (scan=0;scan<128 && scantokey[scan]!=i;scan++) ;
	if (i==' ') scan=57;
/*	if (scan==128) {
		fprintf(stderr,"key: %d ('%c') %s.%d\n",i,i,(stat?"pressed":"released"),scan);
		return 1;
	} */
	kbd_handler(scan,stat);
	for (i=0;i<128;i++)
		if (scanpressed[i]) {
			kbd_handler(i,0);
			scanpressed[i]=0;
		}
	scanpressed[scan]=1;
	return 1;
}

void keyboard_seteventhandler(void *h)
{
	kbd_handler = h;
}

void keyboard_close()
{
	aa_uninitkbd(context);
}
